/* -*- linux-c -*-
 * VLAN		An implementation of 802.1Q VLAN tagging.
 *
 * Version:	0.0.1	03/06/99
 *
 * Authors:	Ben Greear <greearb@candelatech.com>
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 */

#ifndef _LINUX_IF_VLAN_H_
#define _LINUX_IF_VLAN_H_

#ifdef __KERNEL__


/* externally defined structs */
struct vlan_group;
struct device;
struct sk_buff;
struct packet_type;
struct vlan_collection;


#include <linux/proc_fs.h> /* for proc_dir_entry */



/*  Find a VLAN device by the MAC address of it's Ethernet device, and
 *  it's VLAN ID.  The default configuration is to have VLAN's scope
 *  to be box-wide, so the MAC will be ignored.  The mac will only be
 *  looked at if we are configured to have a seperate set of VLANs per
 *  each MAC addressable interface.  Note that this latter option does
 *  NOT follow the spec for VLANs, but may be useful for doing very
 *  large quantities of VLAN MUX/DEMUX onto FrameRelay or ATM PVCs.
 */
struct device *find_802_1Q_vlan_dev(struct device* real_dev,
                                    unsigned short VID); /* vlan.c */


int register_netdevice(struct device *dev); /* found in dev.c */
int unregister_netdevice(struct device *dev); /* found in dev.c */
int dev_new_index(void); /* dev.c */

/* found in vlan_dev.c */
struct net_device_stats* vlan_dev_get_stats(struct device* dev);
int vlan_dev_rebuild_header(struct sk_buff *skb);
int vlan_dev_type_trans(struct sk_buff *skb, struct device *dev,
                        struct packet_type* ptype);
int vlan_dev_hard_header(struct sk_buff *skb, struct device *dev,
                         unsigned short type, void *daddr, void *saddr,
                         unsigned len);
int vlan_dev_hard_start_xmit(struct sk_buff *skb, struct device *dev);
int vlan_dev_change_mtu(struct device *dev, int new_mtu);
int vlan_dev_set_mac_address(struct device *dev, void* addr);
int vlan_dev_open(struct device* dev);
int vlan_dev_stop(struct device* dev);
int vlan_dev_init(struct device* dev);
void vlan_dev_destruct(struct device* dev);
int vlan_dev_set_vlan_flag(char* dev_name, __u32 flag, short flag_val);
/*  I'm ignorant of these right now. --BLG
int vlan_dev_header_cache(struct neighbour *neigh, struct hh_cache *hh);
void vlan_dev_header_cache_update(struct hh_cache *hh, struct device *dev,
                                  unsigned char * haddr);
*/
void vlan_dev_copy_and_sum(struct sk_buff *dest, unsigned char *src,
                           int length, int base);
int vlan_dev_set_ingress_priority(char* dev_name, __u32 skb_prio, short vlan_prio);
int vlan_dev_set_egress_priority(char* dev_name, __u32 skb_prio, short vlan_prio);

/* VLAN multicast stuff */
/* Delete all of the MC list entries from this vlan device.  Also deals
 * with the underlying device...
 */
void vlan_flush_mc_list(struct device* dev);
/* copy the mc_list into the vlan_info structure. */
void vlan_copy_mc_list(struct dev_mc_list* mc_list, struct vlan_dev_info* vlan_info);
/** dmi is a single entry into a dev_mc_list, a single node.  mc_list is
 *  an entire list, and we'll iterate through it.
 */
int vlan_should_add_mc(struct dev_mc_list *dmi, struct dev_mc_list *mc_list);
/** Taken from Gleb + Lennert's VLAN code, and modified... */
void vlan_dev_set_multicast_list(struct device *vlan_dev);


int vlan_collection_add_vlan(struct vlan_collection* vc, unsigned short vlan_id,
                             unsigned short flags);
int vlan_collection_remove_vlan(struct vlan_collection* vc,
                                struct device* vlan_dev);
int vlan_collection_remove_vlan_id(struct vlan_collection* vc, unsigned short vlan_id);



/* found in vlan.c */
/* Our listing of VLAN group(s) */
extern struct vlan_group* p802_1Q_vlan_list;


#define VLAN_NAME "vlan"

/* if this changes, algorithm will have to be reworked because this
 * depends on completely exhausting the VLAN identifier space.  Thus
 * it gives constant time lookup, but it many cases it wastes memory.
 */
#define VLAN_GROUP_ARRAY_LEN 4096

struct vlan_group {
        int real_dev_ifindex; /* The index of the ethernet(like?) device the vlan is attached to. */
        struct device* vlan_devices[VLAN_GROUP_ARRAY_LEN];
   
        struct vlan_group* next; /* the next in the list */
};


/* __Flags__ relating to the vlan ports */
#define VLAN_FLAG_ALLOW_802_3   1
#define VLAN_FLAG_ALLOW_802_1Q  2
#define VLAN_FLAG_IS_IN_USE     4


struct vlan_priority_tci_mapping {
        unsigned long priority;
        unsigned short vlan_qos; /* This should be shifted when first set, so we only do it
                                  * at provisioning time.
                                  * ((skb->priority << 13) & 0xE000)
                                  */
        struct vlan_priority_tci_mapping* next;
};

/* Holds information that makes sense if this device is a VLAN device. */
struct vlan_dev_info {
        /** This will be the mapping that correlates skb->priority to
         * 3 bits of VLAN QOS tags...
         */
        unsigned long ingress_priority_map[8];
        struct vlan_priority_tci_mapping* egress_priority_map[16]; /* hash table */

        unsigned short vlan_id;        /*  The VLAN Identifier for this interface. */
        unsigned short flags;          /* (1 << 0) re_order_header   This option will cause the
                                        *   VLAN code to move around the ethernet header on
                                        *   ingress to make the skb look **exactly** like it
                                        *   came in from an ethernet port.  This destroys some of
                                        *   the VLAN information in the skb, but it fixes programs
                                        *   like DHCP that use packet-filtering and don't understand
                                        *   802.1Q
                                        */
        struct dev_mc_list* old_mc_list;  /* old multi-cast list for the VLAN interface..
                                           * we save this so we can tell what changes were
                                           * made, in order to feed the right changes down
                                           * to the real hardware...
                                           */
        int old_allmulti;               /* similar to above. */
        int old_promiscuity;            /* similar to above. */
        struct device* real_dev;        /* the underlying device/interface */
        struct proc_dir_entry dent;    /* Holds the proc data */
        unsigned long cnt_inc_headroom_on_tx; /* How many times did we have to grow the skb on TX. */
        unsigned long cnt_encap_on_xmit;      /* How many times did we have to encapsulate the skb on TX. */
};

static inline unsigned short vlan_dev_get_egress_qos_mask(struct device* dev, struct sk_buff* skb) {
        struct vlan_priority_tci_mapping* mp = dev->vlan_dev->egress_priority_map[(skb->priority & 0xF)];
        while (mp) {
                if (mp->priority == skb->priority) {
                        return mp->vlan_qos; /* This should already be shifted to mask correctly with
                                              * the VLAN's TCI
                                              */
                }
                mp = mp->next;
        }
        return 0;
}

static inline int vlan_dmi_equals(struct dev_mc_list *dmi1,
                                  struct dev_mc_list *dmi2) {
        return ((dmi1->dmi_addrlen == dmi2->dmi_addrlen) &&
                (memcmp(dmi1->dmi_addr, dmi2->dmi_addr, dmi1->dmi_addrlen) == 0));
}

static inline void vlan_destroy_mc_list(struct dev_mc_list *mc_list) {
        struct dev_mc_list *dmi = mc_list, *next;

        while(dmi) {
                next = dmi->next;
                kfree(dmi);
                dmi = next;
        }
}

#endif /* __KERNEL__ */

/**  These are the IOCTLs relating to the /proc/net/vlan/ * files.
 * Not all may be supported at this time, and some may be primarily
 * used for testing and obtaining non-standard access to kernel
 * devices.
 */

#define VLAN_IOCTL 0x52 /* TODO:  Can I just make these up??? */

enum vlan_ioctls {
        ADD_VLAN_IOCTL = (VLAN_IOCTL << 8),
        DEL_VLAN_IOCTL,
        SET_INGRESS_PRIORITY_IOCTL,
        SET_EGRESS_PRIORITY_IOCTL,
        GET_INGRESS_PRIORITY_IOCTL,
        GET_EGRESS_PRIORITY_IOCTL,
        SET_NAME_TYPE_IOCTL,
        SET_VLAN_FLAG_IOCTL
}; /* vlan_ioctl enum */

enum vlan_name_types {
        VLAN_NAME_TYPE_PLUS_VID, /* Name will look like:  vlan0005 */
        VLAN_NAME_TYPE_RAW_PLUS_VID, /* name will look like:  eth1.0005 */
        VLAN_NAME_TYPE_PLUS_VID_NO_PAD, /* Name will look like:  vlan5 */
        VLAN_NAME_TYPE_RAW_PLUS_VID_NO_PAD, /* Name will look like:  eth0.5 */
        VLAN_NAME_TYPE_HIGHEST
};

struct vlan_ioctl_args {
        char dev1[24];

        union {
                char dev2[24];
                int VID;
                unsigned long skb_priority;
                unsigned long name_type;
                unsigned long bind_type;
                unsigned long flag; /* Matches vlan_dev_info flags */
        } u;

        short vlan_qos;   /* Can also be flag-value, 1 to set, 0 to clear. */
};


#endif
